#include <TROOT.h>
#include <TClass.h>
#include <TGClient.h>
#include <TApplication.h>

#include "MLog.h"
#include "MLogManip.h"

#include "MArgs.h"
#include "MArray.h"
#include "MDirIter.h"

#include "MStatusDisplay.h"

#include "MSequence.h"
#include "MJPedestal.h"
#include "MJCalibration.h"
#include "MJCalibrateSignal.h"

using namespace std;

static void StartUpMessage()
{
    gLog << all << endl;

    //                1         2         3         4         5
    //       12345678901234567890123456789012345678901234567890
    gLog << "========================================================" << endl;
    gLog << "                 Callisto - MARS V" << MARSVER         << endl;
    gLog << "    MARS -- CALibrate LIght Signals and Time Offsets"  << endl;
    gLog << "               Compiled on <" << __DATE__ << ">"       << endl;
    gLog << "                  Using ROOT v" << ROOTVER             << endl;
    gLog << "========================================================" << endl;
    gLog << endl;
}

static void Usage()
{
    //                1         2         3         4         5         6         7         8
    //       12345678901234567890123456789012345678901234567890123456789012345678901234567890
    gLog << all << endl;
    gLog << "Sorry the usage is:" << endl;
    gLog << " callisto [-c] [-y] [options] sequence.txt" << endl << endl;
    gLog << " Arguments:" << endl;
    gLog << "   sequence.txt:             An ascii file defining a sequence of runs" << endl;
    gLog << " Root Options:" << endl;
    gLog << "   -b                        Batch mode (no graphical output to screen)" << endl<<endl;
    gLog << " Operation Modes:" << endl;
    gLog << "   -c                        Calculate the calibration constants" << endl;
    gLog << "   -y                        Extract and calibrate signal" << endl << endl;
    gLog << " Options:" << endl;
    gLog.Usage();
    gLog << endl;
    gLog << "   -f                        Force overwrite of existing files" << endl;
    gLog << "   --in=path                 Path where to search for the data files" << endl;
    gLog << "                             [default=standard path in datacenter]" << endl;
    gLog << "   --iny=path                Path where to search for the calibration files" << endl;
    gLog << "                             [default=local path or output path of Mode-C]" << endl;
    gLog << "   --out=path                Path to write the all results to [def=local path]" << endl;
    gLog << "   --outc=path               Path to write Mode-C result to   [def=local path]" << endl;
    gLog << "   --outy=path               Path to write Mode-Y result to   [def=local path]" << endl;
    gLog << "   --printseq                Print Sequence information" << endl;
    gLog << "   --printfiles              Print Files taken from Sequence" << endl;
    gLog << "   --printonly               Do not excute anything except print" << endl;
    gLog << "   --config=callisto.rc      Resource file [default=callistop.rc]" << endl;
    gLog << endl;
    gLog << "   -?, -h, --help            This help" << endl << endl;
    gLog << " Setup of the two jobs run by callisto (MJPedestal and MJCalibration)" << endl;
    gLog << " can be done with the resource file. See MJPedestal and MJCalibration" << endl;
    gLog << " especially CheckEnv() for more details.   Command line options might" << endl;
    gLog << " be overwritten by the resource file." << endl << endl;
    gLog << "Description:" << endl;
    gLog << " callisto will calculate pedestals  from  pedestal-files defined in a" << endl;
    gLog << " sequence-file.  This pedestals are used to calculate the calibration" << endl;
    gLog << " contants. These constants are stored in a so called calibration-file" << endl;
    gLog << " together with some datacheck plots  which can be viewed using either" << endl;
    gLog << " showplot or MStatusDisplay in the interpreter." << endl;
    gLog << endl;
}

int main(int argc, char **argv)
{
    StartUpMessage();

    //
    // Evaluate arguments
    //
    MArgs arg(argc, argv);

    if (arg.HasOnly("-?") || arg.HasOnly("-h") || arg.HasOnly("--help"))
    {
        Usage();
        return -1;
    }

    gLog.Setup(arg);

    const TString kConfig     = arg.GetStringAndRemove("--config=", "callisto.rc");
    const Bool_t  kPrintSeq   = arg.HasOnlyAndRemove("--printseq");
    const Bool_t  kPrintFiles = arg.HasOnlyAndRemove("--printfiles");
    const Bool_t  kPrintOnly  = arg.HasOnlyAndRemove("--printonly");
    const Bool_t  kOverwrite  = arg.HasOnlyAndRemove("-f");
    const TString kInpathY    = arg.GetStringAndRemove("--iny=",  "");
          TString kInpath     = arg.GetStringAndRemove("--in=",  "");
    const TString kOutpath    = arg.GetStringAndRemove("--out=", "");
          TString kOutpathC   = arg.GetStringAndRemove("--outc=", "");
          TString kOutpathY   = arg.GetStringAndRemove("--outy=", "");

    const Bool_t  kModeC      = arg.HasOnlyAndRemove("-c");
    const Bool_t  kModeY      = arg.HasOnlyAndRemove("-y");

    if (!kModeC && !kModeY)
    {
        gLog << err << "Neither calibration (-c) nor signal extraction (-y) mode specified!" << endl;
        return 0;
    }

    if (!kOutpath.IsNull())
    {
        if (!kOutpathC.IsNull() || !kOutpathY.IsNull())
        {
            gLog << err << "You cannot use --out= and --outc= or --outy= together!" << endl;
            return 0;
        }
        if (kOutpathC.IsNull())
            kOutpathC = kOutpath;
        if (kOutpathY.IsNull())
            kOutpathY = kOutpath;
    }
    if (arg.GetNumOptions()>0)
    {
        gLog << warn << "WARNING - Unknown commandline options..." << endl;
        arg.Print("options");
        gLog << endl;
    }

    //
    // check for the right usage of the program
    //
    if (arg.GetNumArguments()!=1)
    {
        Usage();
        return -1;
    }

    const TString kSequence = arg.GetArgumentStr(0);

    TApplication app("Callisto", &argc, argv);
    if (gROOT->IsBatch() || !gClient)
    {
        gLog << err << "Bombing... maybe your DISPLAY variable is not set correctly!" << endl;
        return 1;
    }

    MSequence seq(kSequence);
    if (kPrintSeq)
    {
        gLog << all;
        gLog.Separator(kSequence);
        seq.Print();
        gLog << endl;
    }
    if (!seq.IsValid())
    {
        gLog << err << "Sequence invalid!" << endl << endl;
        return -1;
    }

    if (kPrintFiles)
    {
        MDirIter Next1, Next2;
        seq.SetupPedRuns(Next1, kInpath);
        seq.SetupDatRuns(Next2, kInpath);

        gLog << all;
        gLog.Separator("Pedestal Files");
        Next1.Print("all");
        gLog << endl;
        gLog.Separator("Data Files");
        Next2.Print("all");
        gLog << endl;
    }

    if (kPrintOnly)
        return 0;

    MArray::Class()->IgnoreTObjectStreamer();
    MParContainer::Class()->IgnoreTObjectStreamer();

    //
    // Update frequency by default = 1Hz
    //
    MStatusDisplay *d = new MStatusDisplay;

    // From now on each 'Exit' means: Terminate the application
    d->SetBit(MStatusDisplay::kExitLoopOnExit);
    d->SetTitle(kSequence);

    if (kModeC)
    {
        MJPedestal job1("MJPedestal1");
        job1.SetSequence(&seq);
        job1.SetEnv(kConfig);
        job1.SetDisplay(d);;
        job1.SetInputPath(kInpath);
        job1.SetOutputPath(kOutpathC);
        job1.SetOverwrite(kOverwrite);

        if (!job1.ProcessFile())
        {
            gLog << err << "Calculation of pedestal failed." << endl << endl;
            return -1;
        }

        if (!job1.GetDisplay())
        {
            gLog << warn << "Display closed by user... execution aborted." << endl << endl;
            return 1;
        }

        MJCalibration job2("MJCalibration");
        job2.SetSequence(&seq);
        job2.SetEnv(kConfig);
        job2.SetDisplay(d);;
        job2.SetBadPixels(job1.GetBadPixels());
        job2.SetInputPath(kInpath);
        job2.SetOutputPath(kOutpathC);
        job2.SetOverwrite(kOverwrite);

        if (!job2.ProcessFile(job1.GetPedestalCam()))
        {
            gLog << err << "Calculation of calibration failed." << endl << endl;
            return -1;
        }

        if (!job2.GetDisplay())
        {
            gLog << warn << "Display closed by user... execution aborted." << endl << endl;
            return 1;
        }
    }

    if (kModeY)
    {
        d->Reset();

        MJCalibrateSignal job("MJCalibrateSignal");
        job.SetDisplay(d);;
        job.SetInputPath(kInpath); // Where to search for data files
        job.SetOutputPath(kOutpathY);
        job.SetOverwrite(kOverwrite);
        // Where to search for calibration files
        if (!job.ProcessFile(seq, kConfig, kModeC ? kOutpathC : kInpathY))
            return -1;

        if (!job.GetDisplay())
        {
            gLog << warn << "Display closed by user... execution aborted." << endl << endl;
            return 1;
        }
    }

    // From now on each 'Close' means: Terminate the application
    d->SetBit(MStatusDisplay::kExitLoopOnClose);

    // Wait until the user decides to exit the application
    app.Run(kFALSE);
    return 0;
}
